#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <unistd.h>
#include <sys/stat.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/time.h>
#include <signal.h>
#include <ctype.h>

#include "mm.hh"
#include "magic.hh"

#include "treetoken.hh"
#include "fileinfo.hh"
#include "infotoken.hh"

const char* database="data";

int dataIncrease=0;

bool gotBreak=false;
bool gotStop=false;

class queryList {
 public:
  queryList() {
    next=0;
    info=0;
  }
//  queryList(queryList* other) {
//    *this=*other;
//  }
  queryList(infoToken* info_) {
    if (info_) {
      info=info_;
      info_=info_->getNext();
      if (info_) {
        next=new queryList(info_);
      } else {
	next=0;
      }
    } else {
      info=0;
      next=0;
    }
  }
  ~queryList() {
    if (next) {
      delete next;
    }
  }
  void attach(queryList* query_) {
    if (info) {
      queryList* trav=this;
      while (trav->next) {
        trav=trav->next;
      } 
      trav->next=query_;
    } else {
      *this=*query_;
    }
  }
  void add(infoToken* info_) {
    if (info) {
      queryList* next_=new queryList;
      next_->next=next;
      next=next_;
      next->setInfo(info_);
    } else {
      info=info_;
    }
  }
  void remove(infoToken* info_) {
    queryList* trav=this;
    while (trav->next) {
      if (trav->next->info==info_) {
	trav->next=trav->next->next;
	break;
      }
      trav=trav->next;
    }
  }
  void setInfo(infoToken* info_) {
    info=info_;
  }
  queryList& operator+=(queryList& other) {
    attach(&other);
    return *this;
  }
  infoToken* info;
  queryList* next;
};

bool
createTree(FILE* file, fileInfo* file, bool findDupes, treeToken* tree, int length=0);

void
countTree(int level, int& wordCount, int& uniqueCount, treeToken& tree);

void						      
showTree(int level, treeToken& tree, char* workSpace=0);

void 
scanDirectory(treeToken& tree, char* path, bool findDupes);

queryList*
findWord(treeToken& tree, char* word, bool ignoreCase=true, int messageLen=-1);

queryList*
findMessage(treeToken& tree, char* message, bool ignoreCase=false, int messageLen=0);

void
sighandler(int signal);

int main(int argc, char** args)
{
  char searchWord[256]={0};
  bool addDirectories=false;
  bool o_showTree=false;
  bool o_showCount=false;
  bool o_help=false;
  bool o_anyOptions=false;
  bool o_checkDupes=true;
  bool o_ignoreCase=false;

  signal(SIGTSTP, sighandler);
  signal(SIGINT, sighandler);
  
  treeToken* tree;
  
  char* dataBase="/var/lib/ds.db";
  if (getenv("DEEPSEEK")) {
    dataBase=getenv("DEEPSEEK");
  }

  if (initMM(dataBase, 1024L*1024L*1024L, false)) {
    tree=(treeToken*)(data+4);
  } else {
    printf("Created new database.\n");
    tree=new treeToken;
  }
  
  int c;
  while ((c=getopt(argc, args, "aith?cd"))!=-1) {
    switch (c) {
     case 'a':
      addDirectories=true;
      o_anyOptions=true;
      break;
     case 'i':
      o_ignoreCase=true;
      break;
     case 't':
      o_showTree=true;
      o_anyOptions=true;
      break;
     case 'c':
      o_showCount=true;
      o_anyOptions=true;
      break;
     case 'd':
      o_checkDupes=false;
      o_anyOptions=true;
      break;
     case 'h':
     case '?':
      o_help=true;
      o_anyOptions=true;
      break;
    }
  }
  
  if (o_help || ((!o_anyOptions) && (argc<=1))) {
    printf("usage: %s [-acith] [word] (or) [directory]\n", args[0]);
    printf(" -c shows total word count & unique word count\n");
    printf(" -a adds [directory] to database\n");
    printf(" -i be case insensitive\n");
    printf(" -d disables dupe-check while adding files\n");
    printf(" -t shows word tree\n");
    printf(" -h this helpscreen\n");
    printf("\ndatabase is acquired from environment variable DEEPSEEK\n");
    printf("(defaults to /var/lib/ds.db)\n");
    exit(0);
  }

  if (addDirectories) {
    int curDataSize=dataStored();
    for (int index=optind; index<argc && !gotBreak; index++) {
      char str[80];
      strcpy(str, args[index]);
      if (str[strlen(str)-1]=='/') {
	str[strlen(str)-1]=0;
      }
      printf("Scanning directory %s\n", str);
      scanDirectory(*tree, str, o_checkDupes);
    }
    printf("total %d bytes data read, %d bytes data written.\n", 
	   dataIncrease, dataStored()-curDataSize);
  } else {
    if (!o_anyOptions) {
      searchWord[0]=0;
      for (int index=optind; index<argc; index++) {
	strcat(searchWord, "(");
	strcat(searchWord, args[index]);
	strcat(searchWord, ")");
	if (index+1<argc) {
	  strcat(searchWord, "&");
	}
      }
    }
  }
   
//  createTree(&fileInfo("/usr/doc/HOWTO/mini/BogoMips"), &tree);
  
  if (o_showCount) {
    int total=0, unique=0;
    countTree(0, total, unique, *tree);
    printf("%d words of which %d were unique.\n", total, unique);
  }
  
  if (o_showTree) {
    showTree(0, *tree);		
  }				

  if (searchWord[0]) {
#ifdef debug
    if (strcmp(searchWord, "(testing123)")==0) {
      strcpy(searchWord, "automat*");
    }
#endif
    printf("Search expression: %s\n", searchWord);
    queryList* result=findMessage(*tree, searchWord, o_ignoreCase);
    int n=0;
    while (result) {
      if (result->info) {
        printf("%s on line %d\n",
    	       (char*)result->info->getFile()->fname,
  	       result->info->getLine());
       	n++;
      }
      result=result->next;
    }
    printf("%d occurances.\n", n);
    if (gotBreak) {
      fprintf(stderr, "Program was interrupted.\n");
    }
  }
  doneMM();
}

void
sighandler(int sig)
{
  switch (sig) {
   case SIGTSTP:
    gotStop=true;
    break;
   case SIGINT:
    gotBreak=true;
    break;
  }
  signal(sig, sighandler);
}

struct openFileInfo {
  magicType type;
  FILE* file;
  struct stat status;
  int length;
};

FILE*
openFile(openFileInfo& info, char* str) {
  FILE* f=0;
  struct stat status;
  if ((f=fopen(str, "r"))) {
    fstat(fileno(f), &status);
    info.status=status;
    info.type=fileType(f);
    fclose(f); f=0;
    char st[256];
    switch (info.type) {
     case typeAscii:
      printf("ascii\n");
      f=fopen(str, "r");
      info.length=status.st_size;
      break;
     case typeTroff:
      printf("troff\n");
      strcpy(st, "nroff ");
      strcat(st, str);
      f=popen(st, "r");
      break;
     case typeGzip:
      strcpy(st, "zcat ");
      strcat(st, str);
      f=popen(st, "r");
      switch (fileType(f)) {
       case typeTroff:
	printf("gzipped troff\n");
	pclose(f);
	strcpy(st, "zcat ");
	strcat(st, str);
	strcat(st, " | nroff");
	f=popen(st, "r");
	break;
       case typePostScript:
	printf("gzipped postscript\n");
	pclose(f);
	f=0;
	break;
       case typeAscii:
	printf("gzipped ascii\n");
	pclose(f);
        f=popen(st, "r");
	break;
       default:
	printf("gzipped unknown\n");
      }
      break;
     case typePostScript:
      printf("postscript\n");
      break;
     default:
      printf("unknown\n");
      f=0;
    }
  } else {
    printf("failed to open\n");
  }
  info.file=f;
  return f;
}

void 
closeFile(openFileInfo& info)
{
  switch (info.type) {
   case typeAscii:
    fclose(info.file);
    break;
   case typeGzip:
   case typeTroff:
    pclose(info.file);
    break;
   default:
    ;
  }
}

void
addFile(treeToken& tree, char* path, bool findDupes)
{
  fileInfo* fInfo;
  openFileInfo oInfo;
  FILE* f;
  struct timeval tv;
  gettimeofday(&tv, 0);
  f=openFile(oInfo, path);
  if (f) {
    int oldHeap=*heapPointer;
    fInfo=new fileInfo(path, oInfo.status, tv);
    if (!createTree(f, fInfo, findDupes, &tree, oInfo.length)) {
      *heapPointer=oldHeap;
    }
    closeFile(oInfo);
  }
}

void
scanDirectory(treeToken& tree, char* path, bool findDupes)
{
  DIR* dir;
  int curDataSize=dataStored();
  int curFileSize=dataIncrease;
  if ((dir=opendir(path))) {
    struct dirent* dirEntry;
    while ((dirEntry=readdir(dir)) && !gotBreak) {
      struct stat status;
      char str[255];
      strcpy(str, path);
      strcat(str, "/");
      strcat(str, dirEntry->d_name);
      stat(str, &status);
      if (!(S_ISDIR(status.st_mode))) {
	printf("%s: ", str);
	addFile(tree, str, findDupes);
      }
    }
  } else {
    printf("%s: ", path);
    addFile(tree, path, findDupes);
  }
  printf("%d bytes read, %d bytes stored.\n",
	 dataIncrease-curFileSize,
	 dataStored()-curDataSize);
	 
}

/*
 * should this function be made inrecursive? less stack swindling, more
 * speed..
 * 
 */

void
addToken(treeToken* tree, char* word, fileInfo* info, int line, int offset)
{
  if (!tree->depth) {		      
    tree->setDepth(new treeToken(0, int(0), int(0), int(0)));
  }
  treeToken* base=tree->depth;
  if (!word[0]) {
    base->letter=0;
    base->setInfo(new infoToken(base->getInfo(), info, line, offset));
  } else {
    while (base->letter<word[0]) {
      if (base->skip) {
        if (base->skip->letter>word[0]) break;
        base=base->skip;
      } else {
        base->skip=new treeToken(word[0], (int)0, (int)0, (int)0);
        base=base->skip;
      }
    }
    if (base->letter!=word[0]) { // inserts BEFORE next character
      base->skip=new treeToken(word[0], base->skip, (int)0, (int)0);
      base=base->skip; // only for addToken-call & count++
    }
    base->count++;
    addToken(base, word+1, info, line, offset);
  }
}

bool
createTree(FILE* f, fileInfo* file, bool findDupes, treeToken* tree, int length=0)
{
  char buf[10240];
  char wordBuf[40];
  int wordWr=0;
  int c;
  int bufRead;
  int offset=0;
  int line=0;
  int d=0;
  bool first=true;
  bool isWord=true;
  do {
    bufRead=fread(buf, 1, sizeof(buf), f);
    for (c=0; c<bufRead; c++) {
      if (buf[c]==10) {
	line++;
      }
      if (buf[c]==10 || buf[c]==13 || isspace(buf[c]) || ispunct(buf[c])) {
        buf[c]=32;
      }
      if (buf[c]==8) {
        if (wordWr>0) {
	  wordWr--;
	}
      }
      if (wordWr>=int(sizeof(wordBuf)-1) || iscntrl(buf[c])) {
        isWord=false;
      }
      if (isWord) {
	wordBuf[wordWr]=buf[c];
      }
      if (buf[c]==' ') {
	if (wordWr<int(sizeof(wordBuf)-1) && isWord) {
          wordBuf[wordWr]=0; wordWr=0;
          if (wordBuf[0] && isalpha(wordBuf[0])) {
	    if (strlen(wordBuf)>=3 ||
		(isupper(wordBuf[0]) ||
		 isupper(wordBuf[1]) ||
		 isupper(wordBuf[2]))) {
              d++;
	      if (first && findDupes) {
	        printf("Checking"); fflush(stdout);
	        queryList* find=findWord(*tree, wordBuf);
		printf("..\r"); fflush(stdout);
	        while (find) {
		  if (find->info) {
	            if (strcmp(find->info->getFile()->fname, file->fname)==0) {
		      printf("                        \r");
	              return false;
	            }
		  }
	          find=find->next;
	        }
	        first=false;
	      }
              addToken(tree, wordBuf, file, line, offset+c-strlen(wordBuf));
	    }
          }
	}
	wordWr=0;
	isWord=true;
      } else {
        wordWr++;
      }
    }
    offset+=bufRead; dataIncrease+=bufRead;
    if (length) {
      printf("%d%c/%d lines/%d words\r", int(offset*100/length), '%', line, d);
      fflush(stdout);
    } else {
      printf("%d lines/%d words\r", line, d);
      fflush(stdout);
    }
    if (gotStop) {
      raise(SIGSTOP);
      gotStop=false;
    }
  } while (!feof(f) && !gotBreak);
  wordBuf[wordWr]=0;
  if (wordBuf[0]) {
    addToken(tree, wordBuf, file, line, offset-strlen(wordBuf));
  }
  printf("                        \r");
  return true;
}

void
countTree(int level, int& wordCount, int& uniqueCount, treeToken& tree)
{
  treeToken* base=tree.depth;
  while (base && !gotBreak) {
    if (base->letter) {
      if (base->depth) {
        countTree(level+1, wordCount, uniqueCount, *base);
      }
    } else {
      mmBlock<infoToken> info=base->getInfo();
      if (info) {
        uniqueCount++;
      }
      while (info) {
        wordCount++;
        info=info->getNext();
      }
    }
    base=base->skip;
  }
}

void
showTree(int level, treeToken& tree, char* workSpace)
{
  if (gotStop) {
    raise(SIGSTOP);
    gotStop=false;
  }
  if (!workSpace) {
    workSpace=new char[40];
  }
  char workCopy[40];
  memcpy(workCopy, workSpace, 40);
  treeToken* base=tree.depth;
  bool anyDepth=false;
  while (base) {
    if (base->letter) {
      workSpace[level]=base->letter;
      putchar(base->letter);
      anyDepth=true;
      if (base->depth) {
        showTree(level+1, *base, workSpace);
	if (gotBreak) {
	  if (!level) {
	    delete[] workSpace;
	  }
	  return;
	}
      }
    }
    base=base->skip; 
  }
  int c=0;
  if (!anyDepth) {
    infoToken* info=tree.depth->getInfo();
    while (info) {
      c++;
      info=info->getNext();
    }
    if (c>0) {
      printf("\t%d\n", c);
      for (c=0; c<level; c++) {
        putchar(workCopy[c]);
      }
    }
  }
  if (!level) {
    delete[] workSpace;
  }
}

queryList*
findWord(treeToken& tree, char* word, bool ignoreCase, int messageLen)
{
  if (gotBreak) {
    return new queryList;
  }
  
  treeToken* base=tree.depth;
  
  if (!messageLen) {
    return new queryList(base->getInfo());
  }
  if (messageLen==-1) {
    messageLen=strlen(word);
  }

  queryList* q=new queryList;

  if (*word=='*') {
    q->attach(new queryList(base->getInfo()));
  }				 

  base=base->skip;

  int dist=1; // perfect hack!
  while (base) {						  
    if ((base->letter==*word) ||
	(ignoreCase && (toupper(base->letter)==toupper(*word))) ||
	(*word=='?') ||
	(*word=='*' && !(dist=0))) {
      queryList* r;
      r=findWord(*base, word+dist, ignoreCase, messageLen-dist);
      if (r->info) {
        q->attach(r);
      } else {
        delete r;
      }
      dist=1;
    }
    base=base->skip;
  }
  return q;
}

queryList*
operate(queryList* list1, queryList* list2, char oper)
{
  queryList* q=new queryList;
  switch (oper) {
    case '&':
    case ' ': {
      queryList* trav1;
      queryList* trav2;
      
      trav1=list1;
      while (trav1 && !gotBreak) {
	trav2=list2;

	infoToken* info1=trav1->info;
	infoToken* info2;

	if (info1) {
          while (trav2 && !gotBreak) {
  	    info2=trav2->info;
	    if (info2) {
	      if ((void*)info1->getFile() == (void*)info2->getFile()) {
	        q->add(info1);
	        break;	
	      }
	    }
	    trav2=trav2->next;
  	  }
	}
        trav1=trav1->next;
      }
    } break;
    case '|': {
      queryList* trav1;
      queryList* trav2;

      trav1=list1;
      while (trav1 && !gotBreak) {
        trav2=list2;
	infoToken* info1=trav1->info;
	infoToken* info2;
	bool same=false;
	
	if (info1) {
	  while (trav2 && !gotBreak) {
	    info2=trav2->info;
	    if (info2) {
	      if ((void*)info1->getFile() == (void*)info2->getFile()) {
		same=true;
	        break;
	      }
	    }
  	    trav2=trav2->next;
 	  }
	  if (!same) {
	    q->add(info1);
	  }
	}
	trav1=trav1->next;
      }

      trav1=list2;
      while (trav1 && !gotBreak) {
        trav2=list1;
	infoToken* info1=trav1->info;
	infoToken* info2;
	bool same=false;
	
	if (info1) {
	  while (trav2 && !gotBreak) {
	    info2=trav2->info;
	    if (info2) {
	      if ((void*)info1->getFile() == (void*)info2->getFile()) {
		same=true;
	        break;
	      }
	    }
  	    trav2=trav2->next;
 	  }
	  if (!same) {
	    q->add(info1);
	  }
	}
	trav1=trav1->next;
      }
    } break;
    case '~': {
      queryList* trav1=list1;
      queryList* trav2;
      
      while (trav1 && !gotBreak) {
        trav2=list2;
	infoToken* info1=trav1->info;
	infoToken* info2;
	
	if (info1) {
	  while (trav2 && !gotBreak) {
	    info2=trav2->info;
	    if (info2) {
	      if ((void*)info1->getFile() == (void*)info2->getFile()) {
		if (abs(info1->getLine()-info2->getLine())<=5) {
		  q->add(info1);
		  break;
		}
	      }
	    }
  	    trav2=trav2->next;
 	  }
	}
	trav1=trav1->next;
      }
    } break;
  }
  return q;
}

void
addMessage(treeToken& tree, 
	   queryList* data[], 
	   int& dataC, 
	   bool ignoreCase, 
	   char* str, 
	   int dataLen,
	   char operand)
{
  if (dataLen>0) {
    data[dataC]=findMessage(tree, str, ignoreCase, dataLen);
  }
  if (dataC) {
    queryList* q=operate(data[0], data[1], operand);
    delete data[0]; data[0]=q;
    delete data[1]; data[1]=0;
  }
  dataC=1;
}

queryList*
findMessage(treeToken& tree, char* message, bool ignoreCase, int messageLen)
{
  if (gotBreak) {
    return new queryList;
  }
  if (!messageLen) {
    messageLen=strlen(message);
  }
  char* first=message;
  queryList* data[2]={0, 0};
  int dataC=0;
  int depth=0;
  char oper=0;

  bool firstPass=true;
  
  for (int c=0; c<messageLen && message[c]; c++) {
    switch (message[c]) {
     case '&':
     case '|':
     case '~':
     case '(':
     case ')':
     case ' ':
      firstPass=false;
      break;
    }
  }
  
  if (firstPass) {
    printf("Searching for ");
    fwrite(message, 1, messageLen, stdout);
    printf("\n");
    return findWord(tree, message, ignoreCase, messageLen);
  } else {
    int c=0;
    while (1) {
      if (message[c]==')') {
        depth--;
	if (depth==0) {
	  data[dataC]=findMessage(tree, first, ignoreCase, message+c-first);
  	  first=message+c+1;
	}
      }
      if (depth==0) {
        switch (message[c]) {
         case '&':
	  addMessage(tree, data, dataC, ignoreCase, first, message-first+c, oper);
          oper='&';
	  first=message+c+1;
          break;
         case '~':
	  addMessage(tree, data, dataC, ignoreCase, first, message-first+c, oper);
          oper='~';
	  first=message+c+1;
          break;
         case '|':
	  addMessage(tree, data, dataC, ignoreCase, first, message-first+c, oper);
          oper='|';
	  first=message+c+1;
          break;
         case '(':
         case ')':
  	  break;
	}
      }
      if (message[c]==0 || c==messageLen) {
	addMessage(tree, data, dataC, ignoreCase, first, message-first+c, oper);
	return data[0];
      }
      if (message[c]=='(') {
        depth++;
        if (depth==1) {
          first=message+c+1;
        }
      }
      c++;
    }
  }
}
